package be.iminds.ilabt.jfed.experiment;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.ExperimentSpecificationFileManager;
import be.iminds.ilabt.jfed.espec.filefetcher.FileFetcher;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecDefaults;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.util.ExperimentRestoreInformation;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.git.GitAuthPreferences;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.KeyUtil;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.ObservableMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.stream.Stream;

/**
 * User: twalcari
 * Date: 1/30/2015
 * Time: 13:24
 */
public class Experiment {
    private static final Logger LOG = LoggerFactory.getLogger(Experiment.class);

//    //for debugging
//    private static Duration DEFAULT_MAX_WAIT_UNTIL_READY = Duration.ofSeconds(10);
//
    private static Duration DEFAULT_MAX_WAIT_UNTIL_READY = Duration.ofMinutes(45);

    /*
    Variables with basic information. These fields will always be set in the constructor
     */
    @Nonnull
    private final String name;
    @Nullable
    private final String subAuthName; //may be null
    @Nonnull
    private final List<GeniUrn> initialShareWithUsers;

    @Nullable
    private final RequestRspecSource newRequestRspecSource;

    @Nullable
    private final RequestRspecSource existingRequestRspecSource;

    @Nullable private final ExperimentSpecification experimentSpecification;
    @Nullable private final ESpecBundle experimentSpecificationBundle;
    @Nullable private final ExperimentSpecificationFileManager experimentSpecificationFileManager;
    @Nonnull private final ESpecLogger eSpecLogger;

    /**
     * No sub-second precision allowed (FEDIBBTDEV-738). Must always be truncated to second precision!
     */
    @Nullable
    private final Instant requestedStartTime;

    /**
     * No sub-second precision allowed (FEDIBBTDEV-738). Must always be truncated to second precision!
     */
    @Nullable
    private final Instant requestedEndTime;

    @Nullable
    private Duration checkReadyInterval = null;
    @Nullable
    private Duration maxWaitUntilReady = DEFAULT_MAX_WAIT_UNTIL_READY;

    /**
     * UserSpecs that must be given access to the experiment resources
     */
    @Nullable
    private List<UserSpec> userSpecs;
    private final boolean requestSla;

    private ExperimentRestoreInformation experimentRestoreInformation;

    /*
    Variables which are subject to change troughout the lifecycle of an experiment
     */
    private final ReadOnlyObjectWrapper<Slice> slice = new ReadOnlyObjectWrapper<>();

    private final ReadOnlyObjectWrapper<ExperimentState> experimentState;

    //always synchronize on experimentPartsObservableList when iterating over any of these 2 lists!
    private final List<ExperimentPart> experimentPartsReadOnlyList;
    private final ObservableList<ExperimentPart> experimentPartsObservableList;
// not good enough            FXCollections.synchronizedObservableList(FXCollections.observableArrayList());

    /*
    Listeners
     */
    private final List<ExperimentChangeListener> experimentChangeListeners = new ArrayList<>();

    private final ObservableMap<String, Object> extras = FXCollections.observableHashMap();

    private final ObservableMap<DistributeSshKeypair, KeyPair> keypairs = FXCollections.observableHashMap();

    /**
     * Constructor used when starting a new experiment in a new to-be-created slice
     */
    public Experiment(@Nonnull String name, @Nullable String subAuthName, @Nonnull List<GeniUrn> initialShareWithUsers,
                      @Nonnull RequestRspecSource newRequestRspecSource,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      @Nullable List<UserSpec> userSpecs, boolean requestSla) {
        assert name != null;
        assert newRequestRspecSource != null;
        assert requestedEndTime != null;
        this.name = name;
        this.subAuthName = subAuthName;
        this.initialShareWithUsers = Collections.unmodifiableList(new ArrayList<>(initialShareWithUsers));
        this.userSpecs = userSpecs;
        this.requestSla = requestSla;

        this.experimentSpecification = null;
        this.experimentSpecificationBundle = null;
        this.experimentSpecificationFileManager = null;
        this.eSpecLogger = new ESpecLogger();

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.newRequestRspecSource = newRequestRspecSource;
        this.existingRequestRspecSource = null;

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.PENDING);

        this.experimentPartsReadOnlyList = new CopyOnWriteArrayList<>();
        this.experimentPartsObservableList = FXCollections.observableList(this.experimentPartsReadOnlyList);

        finishConstruction();
    }

    /**
     * Constructor used when starting a new experiment in a new to-be-created slice, using an ExperimentSpecification
     */
    public Experiment(@Nonnull String name, @Nullable String subAuthName, @Nonnull List<GeniUrn> initialShareWithUsers,
                      @Nonnull ESpecBundle experimentSpecificationBundle,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      @Nullable List<UserSpec> userSpecs, boolean requestSla,
                      @Nullable RSpecGeneratorFactory rSpecGeneratorFactory,
                      @Nonnull GitAuthPreferences gitAuthPreferences) {
        assert name != null;
        assert requestedEndTime != null;
        this.name = name;
        this.subAuthName = subAuthName;
        this.initialShareWithUsers = Collections.unmodifiableList(new ArrayList<>(initialShareWithUsers));
        this.userSpecs = userSpecs;
        this.requestSla = requestSla;

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

//        @SuppressWarnings("ConstantConditions")
//        String userPem = user == null || user.getPrivateKey() == null ? null : new String(KeyUtil.privateKeyToAnyPem(user.getPrivateKey()));

        this.experimentSpecificationBundle = experimentSpecificationBundle;
        ExperimentSpecification experimentSpecification = experimentSpecificationBundle.getExperimentSpecification();
        String requestRspec;
        try {
            requestRspec = new String(new FileFetcher(
                    experimentSpecification.getRspecs().get(0).getSource(),
                    experimentSpecificationBundle,
                    rSpecGeneratorFactory, gitAuthPreferences).fetchBytes(),
                    StandardCharsets.UTF_8);
        } catch (IOException e) {
            throw new RuntimeException("Failed to get Request RSpec from Experiment Specification");
        }
        if (experimentSpecification.getAnsible() != null &&
                experimentSpecification.getAnsible().getAnsibleHostSpec().getHostType() == AnsibleHostSpec.HostType.ADD) {
            this.newRequestRspecSource = createRspecWithAnsibleNode(requestRspec, experimentSpecification.getAnsible());
        } else {
            this.newRequestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
        }
        this.existingRequestRspecSource = null;
        //note: later in the flow ExperimentControllerImpl.allocateResources() will report newRequestRspecSource to ESpecLogger

        if (ESpecLogic.hasAnsibleNode(experimentSpecification)) {
            boolean isWallAnsibleNode = false;
            for (BasicStringRspec.BasicNodeInfo nodeInfo : this.newRequestRspecSource.getBasicNodeInfo()) {
                if (nodeInfo.getClientId() != null &&
                        ESpecLogic.isAnsibleNode(nodeInfo.getClientId(), experimentSpecification)) {
                    GeniUrn compMan = GeniUrn.parse(nodeInfo.getComponentManagerId());
                    isWallAnsibleNode = compMan != null
                            && compMan.getTopLevelAuthority().contains("wall")
                            && compMan.getTopLevelAuthority().contains("ilabt.iminds.be");
                }
            }
            experimentSpecification = isWallAnsibleNode ?
                    addEnableIlabtInternetToESpecAnsible(experimentSpecification) :
                    experimentSpecification;
        }

        this.experimentSpecification = experimentSpecification;

        this.eSpecLogger = new ESpecLogger();
        this.experimentSpecificationFileManager = new ExperimentSpecificationFileManager(
                experimentSpecification, experimentSpecificationBundle, eSpecLogger, rSpecGeneratorFactory,
                gitAuthPreferences);

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.PENDING);

        this.experimentPartsReadOnlyList = new CopyOnWriteArrayList<>();
        this.experimentPartsObservableList = FXCollections.observableList(this.experimentPartsReadOnlyList);

        finishConstruction();
    }

    private ExperimentSpecification addEnableIlabtInternetToESpecAnsible(ExperimentSpecification experimentSpecification) {
        AnsibleSpec origAnsibleSpec = experimentSpecification.getAnsible();
        if (origAnsibleSpec == null)
            return experimentSpecification;

        ExperimentSpecificationBuilder builder = new BasicExperimentSpecificationBuilder(experimentSpecification);
        AnsibleHostSpec origAnsibleHostSpec = origAnsibleSpec.getAnsibleHostSpec();
        if (origAnsibleHostSpec.getExecuteLists().stream()
                .map(ExecuteSpec::getDesc)
                .anyMatch(d -> d.contains("enable-internet"))) {
            return experimentSpecification;
        }
        List<ExecuteSpec> newExecuteList = new ArrayList<>(origAnsibleHostSpec.getExecuteLists());
        newExecuteList.add(new ExecuteSpec(
                new FileSource(
                        FileSource.SourceType.DOWNLOAD,
                        ESpecDefaults.ENABLE_ILABT_INTERNET_URL,
                        ESpecDefaults.ENABLE_ILABT_INTERNET_BASENAME),
                null, null, ESpecDefaults.ENABLE_ILABT_INTERNET_PERMISSIONS,
                true,
                null,
                null));
        AnsibleHostSpec newAnsibleHostSpec = new AnsibleHostSpec(
                origAnsibleHostSpec.getHostType(),
                origAnsibleHostSpec.getNodeName(),
                origAnsibleHostSpec.getUploadLists(),
                newExecuteList,
                origAnsibleHostSpec.getGalaxyCommand(),
                origAnsibleHostSpec.getPlaybookCommand()
        );
        AnsibleSpec newAnsibleSpec = new AnsibleSpec(
                newAnsibleHostSpec,
                origAnsibleSpec.getAnsibleGalaxySpecs(),
                origAnsibleSpec.getAnsiblePlaybookSpecs(),
                origAnsibleSpec.getGroups());
        builder.setAnsible(newAnsibleSpec);
        return builder.create();
    }

    private RequestRspecSource createRspecWithAnsibleNode(@Nonnull String requestRspec, @Nonnull AnsibleSpec ansibleSpec) {
        RequestRspecSource providedRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
        ModelRspec extendedRspecModel = providedRspecSource.getModelRspec();
        if (extendedRspecModel == null) {
            throw new RuntimeException("Failed to get Request RSpec Model from Experiment Specification");
        }
        if (extendedRspecModel.getNodes().isEmpty()) {
            throw new RuntimeException("Cannot add ansible node to RSpec without any nodes");
        }

        RspecNode firstNode = extendedRspecModel.getNodes().get(0);

        RspecFactory factory = RspecFactoryFactory.getRspecFactoryInstance(extendedRspecModel.getModelRspecType());
        RspecNode ansibleNode = factory.createNode(extendedRspecModel);
        if (ansibleSpec.getAnsibleHostSpec().getNodeName() == null) {
            ansibleNode.setClientId("ansible");
        } else {
            ansibleNode.setClientId(ansibleSpec.getAnsibleHostSpec().getNodeName());
        }
        ansibleNode.setComponentManagerId(firstNode.getComponentManagerId());
        ansibleNode.setExclusive(firstNode.getExclusive());
        ansibleNode.setSliverTypeName(firstNode.getSliverTypeName());

        extendedRspecModel.addNode(ansibleNode);
        return new RequestRspecSource(extendedRspecModel);
    }


    /**
     * Constructor used when starting a new experiment in an existing slice
     */
    public Experiment(@Nonnull Slice slice, @Nonnull RequestRspecSource newRequestRspecSource,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      @Nullable List<UserSpec> userSpecs, @Nullable RequestRspecSource existingRequestRspecSource, boolean requestSla) {
        this.requestSla = requestSla;
        this.name = slice.getName();
        this.subAuthName = slice.getSubAuthorityName();
        this.initialShareWithUsers = Collections.emptyList();

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.newRequestRspecSource = newRequestRspecSource;
        this.existingRequestRspecSource = existingRequestRspecSource;

        this.experimentSpecification = null;
        this.experimentSpecificationBundle = null;
        this.experimentSpecificationFileManager = null;
        this.eSpecLogger = new ESpecLogger();

        this.userSpecs = userSpecs;

        this.experimentPartsReadOnlyList = new CopyOnWriteArrayList<>();
        this.experimentPartsObservableList = FXCollections.observableList(this.experimentPartsReadOnlyList);

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.RESTORING);
        registerSlice(slice);

        finishConstruction();
    }


    /**
     * Constructor used when restoring a slice.
     */
    public Experiment(@Nonnull Slice slice, @Nullable Instant requestedStartTime) {
        this.name = slice.getName();
        this.subAuthName = slice.getSubAuthorityName();
        this.initialShareWithUsers = Collections.emptyList();


        this.experimentSpecification = null;
        this.experimentSpecificationBundle = null;
        this.experimentSpecificationFileManager = null;
        this.eSpecLogger = new ESpecLogger();
        this.newRequestRspecSource = null;

        RequestRspecSource existingRequestRspecSource = slice.getRequestRspec();
        if (existingRequestRspecSource == null && slice.getManifestRspec() != null && slice.getManifestRspec().getStringRspec() != null) {
            existingRequestRspecSource =
                    new RequestRspecSource(slice.getManifestRspec().getStringRspec(), ModelRspecType.FX);
        }
        this.existingRequestRspecSource = existingRequestRspecSource;

        this.requestedStartTime = requestedStartTime; //to be fixed slice.getStartDate();
        this.requestedEndTime = null;

        this.experimentPartsReadOnlyList = new CopyOnWriteArrayList<>();
        this.experimentPartsObservableList = FXCollections.observableList(this.experimentPartsReadOnlyList);

        this.userSpecs = null;
        this.requestSla = false;
        registerSlice(slice);

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.RESTORING);

        finishConstruction();
    }

    //instance initializer
    private void finishConstruction() {
        this.experimentState.addListener((observable, oldValue, newValue) -> experimentChangeListeners.forEach(ecl -> ecl.onExperimentStateChange(newValue)));

        this.experimentPartsObservableList.addListener((ListChangeListener<ExperimentPart>) c -> {
            while (c.next()) {
                assert c.wasAdded() : "ExperimentParts should only be added!";
                for (ExperimentPart part : c.getAddedSubList()) {
                    experimentChangeListeners.forEach(ecl -> ecl.onExperimentPartAdded(part));
                }
            }
        });
    }

    public void registerSlice(Slice slice) {
        if (this.slice.get() != null)
            throw new IllegalStateException("Slice is already set in Experiment");
        if (slice == null)
            throw new IllegalArgumentException("Slice is null");
        if (slice.getUrn() == null)
            throw new IllegalArgumentException("Slice has no URN");
        this.slice.setValue(slice);

        this.experimentRestoreInformation = new ExperimentRestoreInformation(slice.getUrn());
        experimentRestoreInformation.bindToExperiment(this);
    }

    public void updateUserSpecs(@Nonnull List<UserSpec> userSpecs){
        this.userSpecs = userSpecs;
    }

    /**
     * Set the maximum time to wait after starting an experiment, until it becomes is ready.
     * If the experiment takes longer, the wait is ended, and the experiment is caonsidered failed.
     *
     * @param maxWaitUntilReady the maximum time to wait after starting an experiment, until it becomes is ready. This may be null, in which case there is no time limit.
     */
    public void setMaxWaitUntilReady(@Nullable Duration maxWaitUntilReady) {
        this.maxWaitUntilReady = maxWaitUntilReady;
    }

    /**
     * Set the time to wait between checking the status of an experiment.
     *
     * @param checkReadyInterval the time to wait between checking the status of an experiment. This may be null, in which case an internal default is used.
     */
    public void setCheckReadyInterval(@Nullable Duration checkReadyInterval) {
        this.checkReadyInterval = checkReadyInterval;
    }

    public void addExperimentChangeListener(ExperimentChangeListener experimentChangeListener) {
        assert (!this.experimentChangeListeners.contains(experimentChangeListener));

        this.experimentChangeListeners.add(experimentChangeListener);
    }

    public void removeExperimentChangeListener(ExperimentChangeListener experimentChangeListener) {
        assert (this.experimentChangeListeners.contains(experimentChangeListener));

        this.experimentChangeListeners.remove(experimentChangeListener);
    }


    @Nonnull
    public String getName() {
        return name;
    }

    @Nullable
    public List<UserSpec> getUserSpecs() {
        return userSpecs;
    }

    @Nullable
    public Duration getMaxWaitUntilReady() {
        return maxWaitUntilReady;
    }

    @Nullable
    public Duration getCheckReadyInterval() {
        return checkReadyInterval;
    }

    public boolean isRequestSla() {
        return requestSla;
    }

    @Nullable
    public Instant getRequestedEndTime() {
        return requestedEndTime;
    }

    @Nullable
    public Instant getRequestedStartTime() {
        return requestedStartTime;
    }

    @Nullable
    public RequestRspecSource getNewRequestRspecSource() {
        return newRequestRspecSource;
    }

    @Nullable
    public ExperimentSpecification getExperimentSpecification() {
        return experimentSpecification;
    }

    @Nullable
    public ExperimentSpecificationFileManager getExperimentSpecificationFileManager() {
        return experimentSpecificationFileManager;
    }

    @Nonnull
    public ESpecLogger getExperimentSpecificationLogger() {
        return eSpecLogger;
    }
    public void addESpecLogListener(@Nonnull ESpecLogListener listener) {
        eSpecLogger.addESpecLogListener(listener);
    }
    public void removeESpecLogListener(@Nonnull ESpecLogListener listener) {
        eSpecLogger.removeESpecLogListener(listener);
    }

    @Nullable
    public ESpecBundle getExperimentSpecificationBundle() {
        return experimentSpecificationBundle;
    }

    @Nullable
    public Slice getSliceOrNull() {
        return slice.get();
    }

    public void requireSlice() {
        if (slice.get() == null)
            throw new IllegalStateException("Experiment should have a slice");
    }

    @Nonnull
    public Optional<Slice> getSliceOpt() {
        return Optional.ofNullable(slice.get());
    }

    /**
     * This returns the slice related to this experiment, or throws an exception if there is none.
     *
     * @return the slice related to this experiment (never null)
     */
    @Nonnull
    public Slice getSlice() {
        Slice res = slice.get();
        if (res == null) {
            throw new IllegalStateException("The Experiment does not have a slice");
        }
        return res;
    }

    @Nullable
    public GeniUrn getSliceUrn() {
        return slice.get() == null ? null : slice.get().getUrn();
    }

    @Nullable
    public String getSliceUrnString() {
        return slice.get() == null ? null : slice.get().getUrnString();
    }

    public ReadOnlyObjectProperty<Slice> sliceProperty() {
        return slice.getReadOnlyProperty();
    }

    @Nonnull
    public ExperimentState getExperimentState() {
        return experimentState.get();
    }

    @Nonnull
    public ReadOnlyObjectProperty<ExperimentState> experimentStateProperty() {
        return experimentState.getReadOnlyProperty();
    }

    @Nullable
    public String getSubAuthName() {
        return subAuthName;
    }

    @Nonnull
    public List<GeniUrn> getInitialShareWithUsers() {
        return initialShareWithUsers;
    }

    public boolean isReservationMade() {
        synchronized (experimentPartsObservableList) {
            return experimentPartsReadOnlyList.stream()
                    .anyMatch(experimentPart -> experimentPart.getState() == InternalState.RESERVED);
        }
    }

    public boolean getAnyPartSawNotEnoughFreeResourcesError() {
        synchronized (experimentPartsObservableList) {
            return experimentPartsReadOnlyList.stream()
                    .anyMatch(ExperimentPart::getSawNotEnoughFreeResourcesError);
        }
    }

    public ExperimentRestoreInformation getExperimentRestoreInformation() {
        return experimentRestoreInformation;
    }

    public void setExperimentState(@Nonnull ExperimentState experimentState) {
        LOG.debug("Setting state for experiment {} to {}", name, experimentState);
        this.experimentState.set(experimentState);

    }

    @Nullable
    public RequestRspecSource getExistingRequestRspecSource() {
        return existingRequestRspecSource;
    }

    @Nonnull
    public ObservableList<ExperimentPart> getPartsObservableList() {
        synchronized (experimentPartsObservableList) {
            return experimentPartsObservableList;
        }
    }
    @Nonnull
    public List<ExperimentPart> getPartsListCopy() {
        synchronized (experimentPartsObservableList) {
            return new ArrayList<>(experimentPartsReadOnlyList);
        }
    }
    @Nonnull
    public Stream<ExperimentPart> getPartsStream() {
        synchronized (experimentPartsObservableList) {
            return new ArrayList<>(experimentPartsReadOnlyList).stream();
        }
    }
    public int getPartsSize() {
        return experimentPartsReadOnlyList.size();
    }
    public boolean isPartsEmpty() {
        return experimentPartsReadOnlyList.isEmpty();
    }

    public void addExperimentPart(@Nonnull ExperimentPart experimentPart) {
        //noinspection ObjectEquality
        assert experimentPart.getExperiment() == this;

        synchronized (experimentPartsObservableList) {
            this.experimentPartsObservableList.add(experimentPart);
        }
    }

    @Nullable
    public SfaExperimentPart getPart(Server auth) {
        synchronized (experimentPartsObservableList) {
            return experimentPartsReadOnlyList.stream()
                    .filter(part -> part instanceof SfaExperimentPart &&
                            Objects.equals(((SfaExperimentPart) part).getConnectSfaAuthority(), auth))
                    .map(SfaExperimentPart.class::cast)
                    .findFirst().orElse(null);
        }
    }

    @Nullable
    public ExperimentPart getPart(Sliver sliver) {
        synchronized (experimentPartsObservableList) {
            return experimentPartsReadOnlyList.stream()
                    .filter(part -> part instanceof SfaExperimentPart
                            && ((SfaExperimentPart) part).getSlivers().contains(sliver))
                    .findFirst().orElse(null);
        }
    }

    @Nullable
    public ExperimentPart getPart(GeniUrn sliverUrn) {
        synchronized (experimentPartsObservableList) {
            for (ExperimentPart experimentPart : experimentPartsReadOnlyList) {
                if (experimentPart instanceof SfaExperimentPart) {
                    SfaExperimentPart sfaExperimentPart = (SfaExperimentPart) experimentPart;

                    for (Sliver sliver : sfaExperimentPart.getSlivers()) {
                        if (Objects.equals(sliver.getUrn(), sliverUrn))
                            return experimentPart;
                    }
                }
            }
        }
        return null;
    }


    public ObservableMap<String, Object> getExtras() {
        return extras;
    }

    public ObservableMap<DistributeSshKeypair, KeyPair> getKeypairs() {
        return keypairs;
    }
}
